"""
Registry mapping evaluator slugs to their request/response Pydantic models.

This enables type-safe validation of inputs and parsing of outputs.

DO NOT EDIT MANUALLY - Regenerate with:
    ./scripts/generate-models.sh /path/to/swagger.json
"""

from typing import Dict, Type, Optional
from pydantic import BaseModel

from .request import (
    AgentEfficiencyRequest,
    AgentFlowQualityRequest,
    AgentGoalAccuracyRequest,
    AgentGoalCompletenessRequest,
    AgentToolErrorDetectorRequest,
    AgentToolTrajectoryRequest,
    AnswerCompletenessRequest,
    AnswerCorrectnessRequest,
    AnswerRelevancyRequest,
    CharCountRatioRequest,
    CharCountRequest,
    ContextRelevanceRequest,
    ConversationQualityRequest,
    FaithfulnessRequest,
    HtmlComparisonRequest,
    InstructionAdherenceRequest,
    IntentChangeRequest,
    JSONValidatorRequest,
    PIIDetectorRequest,
    PerplexityRequest,
    PlaceholderRegexRequest,
    ProfanityDetectorRequest,
    PromptInjectionRequest,
    PromptPerplexityRequest,
    RegexValidatorRequest,
    SQLValidatorRequest,
    SecretsDetectorRequest,
    SemanticSimilarityRequest,
    SexismDetectorRequest,
    ToneDetectionRequest,
    TopicAdherenceRequest,
    ToxicityDetectorRequest,
    UncertaintyDetectorRequest,
    WordCountRatioRequest,
    WordCountRequest,
)

from .response import (
    AgentEfficiencyResponse,
    AgentFlowQualityResponse,
    AgentGoalAccuracyResponse,
    AgentGoalCompletenessResponse,
    AgentToolErrorDetectorResponse,
    AgentToolTrajectoryResponse,
    AnswerCompletenessResponse,
    AnswerCorrectnessResponse,
    AnswerRelevancyResponse,
    CharCountRatioResponse,
    CharCountResponse,
    ContextRelevanceResponse,
    ConversationQualityResponse,
    FaithfulnessResponse,
    HtmlComparisonResponse,
    InstructionAdherenceResponse,
    IntentChangeResponse,
    JSONValidatorResponse,
    PIIDetectorResponse,
    PerplexityResponse,
    PlaceholderRegexResponse,
    ProfanityDetectorResponse,
    PromptInjectionResponse,
    PromptPerplexityResponse,
    RegexValidatorResponse,
    SQLValidatorResponse,
    SecretsDetectorResponse,
    SemanticSimilarityResponse,
    SexismDetectorResponse,
    ToneDetectionResponse,
    TopicAdherenceResponse,
    ToxicityDetectorResponse,
    UncertaintyDetectorResponse,
    WordCountRatioResponse,
    WordCountResponse,
)


# Mapping from evaluator slug to request model
REQUEST_MODELS: Dict[str, Type[BaseModel]] = {
    "agent-efficiency": AgentEfficiencyRequest,
    "agent-flow-quality": AgentFlowQualityRequest,
    "agent-goal-accuracy": AgentGoalAccuracyRequest,
    "agent-goal-completeness": AgentGoalCompletenessRequest,
    "agent-tool-error-detector": AgentToolErrorDetectorRequest,
    "agent-tool-trajectory": AgentToolTrajectoryRequest,
    "answer-completeness": AnswerCompletenessRequest,
    "answer-correctness": AnswerCorrectnessRequest,
    "answer-relevancy": AnswerRelevancyRequest,
    "char-count": CharCountRequest,
    "char-count-ratio": CharCountRatioRequest,
    "context-relevance": ContextRelevanceRequest,
    "conversation-quality": ConversationQualityRequest,
    "faithfulness": FaithfulnessRequest,
    "html-comparison": HtmlComparisonRequest,
    "instruction-adherence": InstructionAdherenceRequest,
    "intent-change": IntentChangeRequest,
    "json-validator": JSONValidatorRequest,
    "perplexity": PerplexityRequest,
    "pii-detector": PIIDetectorRequest,
    "placeholder-regex": PlaceholderRegexRequest,
    "profanity-detector": ProfanityDetectorRequest,
    "prompt-injection": PromptInjectionRequest,
    "prompt-perplexity": PromptPerplexityRequest,
    "regex-validator": RegexValidatorRequest,
    "secrets-detector": SecretsDetectorRequest,
    "semantic-similarity": SemanticSimilarityRequest,
    "sexism-detector": SexismDetectorRequest,
    "sql-validator": SQLValidatorRequest,
    "tone-detection": ToneDetectionRequest,
    "topic-adherence": TopicAdherenceRequest,
    "toxicity-detector": ToxicityDetectorRequest,
    "uncertainty-detector": UncertaintyDetectorRequest,
    "word-count": WordCountRequest,
    "word-count-ratio": WordCountRatioRequest,
}

# Mapping from evaluator slug to response model
RESPONSE_MODELS: Dict[str, Type[BaseModel]] = {
    "agent-efficiency": AgentEfficiencyResponse,
    "agent-flow-quality": AgentFlowQualityResponse,
    "agent-goal-accuracy": AgentGoalAccuracyResponse,
    "agent-goal-completeness": AgentGoalCompletenessResponse,
    "agent-tool-error-detector": AgentToolErrorDetectorResponse,
    "agent-tool-trajectory": AgentToolTrajectoryResponse,
    "answer-completeness": AnswerCompletenessResponse,
    "answer-correctness": AnswerCorrectnessResponse,
    "answer-relevancy": AnswerRelevancyResponse,
    "char-count": CharCountResponse,
    "char-count-ratio": CharCountRatioResponse,
    "context-relevance": ContextRelevanceResponse,
    "conversation-quality": ConversationQualityResponse,
    "faithfulness": FaithfulnessResponse,
    "html-comparison": HtmlComparisonResponse,
    "instruction-adherence": InstructionAdherenceResponse,
    "intent-change": IntentChangeResponse,
    "json-validator": JSONValidatorResponse,
    "perplexity": PerplexityResponse,
    "pii-detector": PIIDetectorResponse,
    "placeholder-regex": PlaceholderRegexResponse,
    "profanity-detector": ProfanityDetectorResponse,
    "prompt-injection": PromptInjectionResponse,
    "prompt-perplexity": PromptPerplexityResponse,
    "regex-validator": RegexValidatorResponse,
    "secrets-detector": SecretsDetectorResponse,
    "semantic-similarity": SemanticSimilarityResponse,
    "sexism-detector": SexismDetectorResponse,
    "sql-validator": SQLValidatorResponse,
    "tone-detection": ToneDetectionResponse,
    "topic-adherence": TopicAdherenceResponse,
    "toxicity-detector": ToxicityDetectorResponse,
    "uncertainty-detector": UncertaintyDetectorResponse,
    "word-count": WordCountResponse,
    "word-count-ratio": WordCountRatioResponse,
}


def get_request_model(slug: str) -> Optional[Type[BaseModel]]:
    """Get the request model for an evaluator by slug."""
    return REQUEST_MODELS.get(slug)


def get_response_model(slug: str) -> Optional[Type[BaseModel]]:
    """Get the response model for an evaluator by slug."""
    return RESPONSE_MODELS.get(slug)



from typing import Any, Dict, Optional

import requests
from colorama import Fore


class HTTPClient:
    base_url: str
    api_key: str
    version: str

    def __init__(self, base_url: str, api_key: str, version: str):
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self.version = version

    def _headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
            "X-Traceloop-SDK-Version": self.version,
        }

    def post(self, path: str, data: Dict[str, Any]) -> Any:
        """
        Make a POST request to the API
        """
        try:
            response = requests.post(
                f"{self.base_url}/v2/{path.lstrip('/')}",
                json=data,
                headers=self._headers(),
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            error_msg = self._format_error_message(path, e)
            print(Fore.RED + error_msg + Fore.RESET)
            return None

    def get(self, path: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """
        Make a GET request to the API
        """
        try:
            response = requests.get(
                f"{self.base_url}/v2/{path.lstrip('/')}",
                params=params,
                headers=self._headers(),
            )
            response.raise_for_status()

            content_type = response.headers.get("content-type", "").lower()
            if "text/csv" in content_type or "application/x-ndjson" in content_type:
                return response.text
            else:
                return response.json()
        except requests.exceptions.RequestException as e:
            error_msg = self._format_error_message(path, e)
            print(Fore.RED + error_msg + Fore.RESET)
            return None

    def delete(self, path: str) -> bool:
        """
        Make a DELETE request to the API
        """
        try:
            response = requests.delete(
                f"{self.base_url}/v2/{path.lstrip('/')}", headers=self._headers()
            )
            response.raise_for_status()
            return response.status_code == 204 or response.status_code == 200
        except requests.exceptions.RequestException as e:
            error_msg = self._format_error_message(path, e)
            print(Fore.RED + error_msg + Fore.RESET)
            return False

    def put(self, path: str, data: Dict[str, Any]) -> Any:
        """
        Make a PUT request to the API
        """
        try:
            response = requests.put(
                f"{self.base_url}/v2/{path.lstrip('/')}",
                json=data,
                headers=self._headers(),
            )
            response.raise_for_status()
            if response.content:
                return response.json()
            else:
                return {}
        except requests.exceptions.RequestException as e:
            error_msg = self._format_error_message(path, e)
            print(Fore.RED + error_msg + Fore.RESET)
            return None

    def _format_error_message(self, path: str, exception: requests.exceptions.RequestException) -> str:
        """
        Format a detailed error message including server response if available
        """
        error_parts = [f"Error making request to {path}: {str(exception)}"]

        # Try to extract error details from response
        if hasattr(exception, 'response') and exception.response is not None:
            response = exception.response

            # Try to parse JSON error from response
            try:
                error_data = response.json()
                if isinstance(error_data, dict):
                    # Check common error fields
                    if 'error' in error_data:
                        error_parts.append(f"Server error: {error_data['error']}")
                    elif 'message' in error_data:
                        error_parts.append(f"Server message: {error_data['message']}")
                    elif 'msg' in error_data:
                        error_parts.append(f"Server message: {error_data['msg']}")
                    else:
                        # Include the entire JSON if no standard field found
                        error_parts.append(f"Server response: {error_data}")
            except (ValueError, AttributeError):
                # Not JSON, try to get text
                try:
                    error_text = response.text
                    if error_text and len(error_text) < 500:  # Only include short error messages
                        error_parts.append(f"Server response: {error_text}")
                except Exception:
                    pass

        return "\n".join(error_parts)



import pytest
from traceloop.sdk.evaluator.evaluator import (
    validate_and_normalize_task_output,
    _validate_evaluator_input,
)
from traceloop.sdk.evaluator.config import EvaluatorDetails


class TestValidateTaskOutput:
    """Tests for validate_task_output function"""

    def test_validate_task_output_with_no_evaluators(self):
        """Test that validation passes when no evaluators are provided"""
        task_output = {"text": "hello"}
        evaluators = []

        # Should not raise any exception
        validate_and_normalize_task_output(task_output, evaluators)

    def test_validate_task_output_with_evaluators_no_required_fields(self):
        """Test that validation passes when evaluators have no required fields"""
        task_output = {"text": "hello", "score": 0.9}
        evaluators = [
            EvaluatorDetails(slug="evaluator1"),
            EvaluatorDetails(slug="evaluator2", config={"threshold": 0.5}),
        ]

        # Should not raise any exception
        validate_and_normalize_task_output(task_output, evaluators)

    def test_validate_task_output_with_valid_output(self):
        """Test that validation passes when all required fields are present"""
        task_output = {"text": "hello", "prompt": "say hello", "response": "world"}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
            EvaluatorDetails(
                slug="relevance-checker",
                required_input_fields=["prompt", "response"],
            ),
        ]

        # Should not raise any exception
        validate_and_normalize_task_output(task_output, evaluators)

    def test_validate_task_output_missing_single_field(self):
        """Test that validation fails when a single required field is missing"""
        task_output = {"prompt": "say hello"}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "Task output missing required fields for evaluators:" in error_message
        assert "pii-detector requires:" in error_message
        assert "'text'" in error_message
        assert "Task output contains: ['prompt']" in error_message
        assert (
            "Hint: Update your task function to return a dictionary "
            "with the required fields."
        ) in error_message

    def test_validate_task_output_missing_multiple_fields_single_evaluator(self):
        """Test that validation fails when multiple fields are missing for one evaluator"""
        task_output = {"score": 0.9}
        evaluators = [
            EvaluatorDetails(
                slug="relevance-checker",
                required_input_fields=["prompt", "response", "context"],
            ),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "relevance-checker requires:" in error_message
        assert "'context'" in error_message
        assert "'prompt'" in error_message
        assert "'response'" in error_message

    def test_validate_task_output_missing_fields_multiple_evaluators(self):
        """Test that validation fails when fields are missing for multiple evaluators"""
        task_output = {"score": 0.9}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
            EvaluatorDetails(
                slug="relevance-checker",
                required_input_fields=["prompt", "response"],
            ),
            EvaluatorDetails(slug="tone-analyzer", required_input_fields=["text"]),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "pii-detector requires:" in error_message
        assert "relevance-checker requires:" in error_message
        assert "tone-analyzer requires:" in error_message
        assert "'text'" in error_message
        assert "'prompt'" in error_message
        assert "'response'" in error_message

    def test_validate_task_output_partial_match(self):
        """Test validation when some evaluators pass and some fail"""
        task_output = {"text": "hello world"}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
            EvaluatorDetails(
                slug="relevance-checker",
                required_input_fields=["prompt", "response"],
            ),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        # Should only mention the failing evaluator
        assert "relevance-checker requires:" in error_message
        assert "pii-detector requires:" not in error_message

    def test_validate_task_output_empty_task_output(self):
        """Test validation with empty task output"""
        task_output = {}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "Task output contains: []" in error_message

    def test_validate_task_output_with_extra_fields(self):
        """Test that validation passes when task output has extra fields"""
        task_output = {
            "text": "hello",
            "prompt": "say hello",
            "response": "world",
            "extra_field": "value",
            "another_field": 123,
        }
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
        ]

        # Should not raise any exception - extra fields are allowed
        validate_and_normalize_task_output(task_output, evaluators)

    def test_validate_task_output_case_sensitive_field_names(self):
        """Test that field name matching is case-sensitive"""
        task_output = {"Text": "hello"}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "pii-detector requires:" in error_message
        assert "'text'" in error_message
        assert "Task output contains: ['Text']" in error_message

    def test_validate_task_output_with_evaluator_config(self):
        """Test validation with evaluators that have config"""
        task_output = {"text": "hello world"}
        evaluators = [
            EvaluatorDetails(
                slug="pii-detector",
                version="v2",
                config={"probability_threshold": 0.8, "mode": "strict"},
                required_input_fields=["text"],
            ),
        ]

        # Should not raise any exception - config shouldn't affect validation
        validate_and_normalize_task_output(task_output, evaluators)

    def test_validate_task_output_mixed_evaluators(self):
        """Test validation with a mix of evaluators with and without required fields"""
        task_output = {"text": "hello", "score": 0.9}
        evaluators = [
            EvaluatorDetails(slug="evaluator-no-requirements"),
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
            EvaluatorDetails(slug="another-no-requirements", config={"key": "value"}),
            EvaluatorDetails(
                slug="relevance-checker",
                required_input_fields=["prompt"],
            ),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        # Should only mention failing evaluator
        assert "relevance-checker requires:" in error_message
        assert "'prompt'" in error_message
        assert "evaluator-no-requirements" not in error_message
        assert "pii-detector" not in error_message or "pii-detector requires:" not in error_message

    def test_validate_task_output_duplicate_required_fields(self):
        """Test validation when multiple evaluators require the same field"""
        task_output = {"score": 0.9}
        evaluators = [
            EvaluatorDetails(slug="pii-detector", required_input_fields=["text"]),
            EvaluatorDetails(slug="tone-analyzer", required_input_fields=["text"]),
            EvaluatorDetails(
                slug="sentiment-analyzer",
                required_input_fields=["text", "language"],
            ),
        ]

        with pytest.raises(ValueError) as exc_info:
            validate_and_normalize_task_output(task_output, evaluators)

        error_message = str(exc_info.value)
        assert "pii-detector requires:" in error_message
        assert "tone-analyzer requires:" in error_message
        assert "sentiment-analyzer requires:" in error_message


class TestValidateEvaluatorInput:
    """Tests for _validate_evaluator_input function"""

    def test_validate_input_no_request_model(self):
        """Validation passes for unknown slugs (no request model registered)."""
        # Should not raise - unknown slug has no model to validate against
        _validate_evaluator_input("unknown-evaluator", {"text": "hello"})

    def test_validate_input_valid_input_only(self):
        """Validation passes for evaluators that only require input (no config)."""
        _validate_evaluator_input(
            "pii-detector",
            {"text": "Please contact John at john@email.com"},
        )

    def test_validate_input_missing_required_input_field(self):
        """Validation fails when a required input field is missing."""
        with pytest.raises(ValueError, match="Invalid input for 'pii-detector'"):
            _validate_evaluator_input("pii-detector", {"wrong_field": "value"})

    def test_validate_input_with_optional_config(self):
        """Validation passes for evaluators with optional config when config is provided."""
        _validate_evaluator_input(
            "pii-detector",
            {"text": "Some text"},
            evaluator_config={"probability_threshold": 0.8},
        )

    def test_validate_input_with_optional_config_omitted(self):
        """Validation passes for evaluators with optional config when config is omitted."""
        _validate_evaluator_input(
            "toxicity-detector",
            {"text": "Some text"},
        )

    def test_validate_agent_flow_quality_with_required_config(self):
        """Validation passes for agent-flow-quality when config is provided."""
        _validate_evaluator_input(
            "agent-flow-quality",
            {
                "trajectory_completions": '["Found 5 flights"]',
                "trajectory_prompts": '["Search for flights"]',
            },
            evaluator_config={
                "conditions": ["no tools called"],
                "threshold": 0.5,
            },
        )

    def test_validate_agent_flow_quality_missing_config_fails(self):
        """Validation fails for agent-flow-quality when required config is missing."""
        with pytest.raises(ValueError, match="Invalid input for 'agent-flow-quality'"):
            _validate_evaluator_input(
                "agent-flow-quality",
                {
                    "trajectory_completions": '["Found 5 flights"]',
                    "trajectory_prompts": '["Search for flights"]',
                },
            )

    def test_validate_agent_flow_quality_missing_input_fields(self):
        """Validation fails for agent-flow-quality when required input fields are missing."""
        with pytest.raises(ValueError, match="Invalid input for 'agent-flow-quality'"):
            _validate_evaluator_input(
                "agent-flow-quality",
                {"wrong_field": "value"},
                evaluator_config={
                    "conditions": ["no tools called"],
                    "threshold": 0.5,
                },
            )

    def test_validate_agent_goal_completeness_optional_config(self):
        """agent-goal-completeness has optional config - passes with or without it."""
        input_data = {
            "trajectory_completions": '["Account created"]',
            "trajectory_prompts": '["Create new account"]',
        }
        # Without config
        _validate_evaluator_input("agent-goal-completeness", input_data)
        # With config
        _validate_evaluator_input(
            "agent-goal-completeness",
            input_data,
            evaluator_config={"threshold": 0.5},
        )

    def test_validate_agent_tool_trajectory_optional_config(self):
        """agent-tool-trajectory has optional config - passes with or without it."""
        input_data = {
            "executed_tool_calls": '[{"name": "search", "input": {"query": "weather"}}]',
            "expected_tool_calls": '[{"name": "search", "input": {"query": "weather"}}]',
        }
        # Without config
        _validate_evaluator_input("agent-tool-trajectory", input_data)
        # With config
        _validate_evaluator_input(
            "agent-tool-trajectory",
            input_data,
            evaluator_config={"order_sensitive": True, "threshold": 0.5},
        )



import pytest


@pytest.mark.vcr
def test_get_dataset_by_version(datasets):
    csv_data = datasets.get_version_csv(slug="test-qa", version="v1")
    assert isinstance(csv_data, str)


@pytest.mark.vcr
def test_publish_dataset(datasets):
    unique_slug = "test-publish-dataset"

    # Create a simple CSV for the dataset
    import tempfile
    import os

    csv_content = """Name,Value
Test,123"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".csv", delete=False) as f:
        f.write(csv_content)
        f.flush()
        csv_path = f.name

        dataset = datasets.from_csv(
            file_path=csv_path,
            slug=unique_slug,
            name="Test Publish Dataset",
            description="Dataset for testing publish functionality",
        )

        # Try to publish it
        version = dataset.publish()
        assert isinstance(version, str)

    os.unlink(csv_path)



import asyncio
import json
import dataclasses

import pytest
from openai import OpenAI, AsyncOpenAI
from opentelemetry.semconv._incubating.attributes import (
    gen_ai_attributes as GenAIAttributes,
)
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace.status import StatusCode
from traceloop.sdk import Traceloop
from traceloop.sdk.decorators import workflow, task


@pytest.fixture
def openai_client():
    return OpenAI()


@pytest.fixture
def async_openai_client():
    return AsyncOpenAI()


@pytest.mark.vcr
def test_simple_workflow(exporter, openai_client):
    @task(name="something_creator", version=2)
    def create_something(what: str, subject: str):
        Traceloop.set_prompt("Tell me a {what} about {subject}", {"what": what, "subject": subject}, 5)
        completion = openai_client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "user", "content": f"Tell me a {what} about {subject}"}],
        )
        return completion.choices[0].message.content

    @workflow(name="pirate_joke_generator", version=1)
    def joke_workflow():
        return create_something("joke", subject="OpenTelemetry")

    joke = joke_workflow()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "openai.chat",
        "something_creator.task",
        "pirate_joke_generator.workflow",
    ]
    open_ai_span = next(span for span in spans if span.name == "openai.chat")
    assert open_ai_span.attributes[f"{GenAIAttributes.GEN_AI_PROMPT}.0.content"] == "Tell me a joke about OpenTelemetry"
    assert open_ai_span.attributes.get(f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content")
    assert open_ai_span.attributes.get("traceloop.prompt.template") == "Tell me a {what} about {subject}"
    assert open_ai_span.attributes.get("traceloop.prompt.template_variables.what") == "joke"
    assert open_ai_span.attributes.get("traceloop.prompt.template_variables.subject") == "OpenTelemetry"
    assert open_ai_span.attributes.get("traceloop.prompt.version") == "5"

    workflow_span = next(span for span in spans if span.name == "pirate_joke_generator.workflow")
    task_span = next(span for span in spans if span.name == "something_creator.task")
    assert json.loads(task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_INPUT]) == {
        "args": ["joke"],
        "kwargs": {"subject": "OpenTelemetry"},
    }

    assert json.loads(task_span.attributes.get(SpanAttributes.TRACELOOP_ENTITY_OUTPUT)) == joke
    assert task_span.parent.span_id == workflow_span.context.span_id
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "pirate_joke_generator"
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_ENTITY_VERSION] == 1
    assert task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_VERSION] == 2


@pytest.mark.vcr
@pytest.mark.asyncio
async def test_simple_aworkflow(exporter, async_openai_client):
    @task(name="something_creator", version=2)
    async def create_something(what: str, subject: str):
        Traceloop.set_prompt("Tell me a {what} about {subject}", {"what": what, "subject": subject}, 5)
        completion = await async_openai_client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "user", "content": f"Tell me a {what} about {subject}"}],
        )
        return completion.choices[0].message.content

    @workflow(name="pirate_joke_generator", version=1)
    async def joke_workflow():
        return await create_something("joke", subject="OpenTelemetry")

    joke = await joke_workflow()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == [
        "openai.chat",
        "something_creator.task",
        "pirate_joke_generator.workflow",
    ]
    open_ai_span = next(span for span in spans if span.name == "openai.chat")
    assert open_ai_span.attributes[f"{GenAIAttributes.GEN_AI_PROMPT}.0.content"] == "Tell me a joke about OpenTelemetry"
    assert open_ai_span.attributes.get(f"{GenAIAttributes.GEN_AI_COMPLETION}.0.content")
    assert open_ai_span.attributes.get("traceloop.prompt.template") == "Tell me a {what} about {subject}"
    assert open_ai_span.attributes.get("traceloop.prompt.template_variables.what") == "joke"
    assert open_ai_span.attributes.get("traceloop.prompt.template_variables.subject") == "OpenTelemetry"
    assert open_ai_span.attributes.get("traceloop.prompt.version") == "5"

    workflow_span = next(span for span in spans if span.name == "pirate_joke_generator.workflow")
    task_span = next(span for span in spans if span.name == "something_creator.task")
    assert json.loads(task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_INPUT]) == {
        "args": ["joke"],
        "kwargs": {"subject": "OpenTelemetry"},
    }

    assert json.loads(task_span.attributes.get(SpanAttributes.TRACELOOP_ENTITY_OUTPUT)) == joke
    assert task_span.parent.span_id == workflow_span.context.span_id
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "pirate_joke_generator"
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_ENTITY_VERSION] == 1
    assert task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_VERSION] == 2


@pytest.mark.vcr
def test_streaming_workflow(exporter, openai_client):
    @task(name="pirate_joke_generator")
    def joke_task():
        response_stream = openai_client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "user", "content": "Tell me a joke about OpenTelemetry"}],
            stream=True,
        )
        for chunk in response_stream:
            yield chunk

    @task(name="joke_runner")
    def joke_runner():
        res = joke_task()
        return res

    @workflow(name="joke_manager")
    def joke_workflow():
        res = joke_runner()
        for chunk in res:
            pass

    joke_workflow()

    spans = exporter.get_finished_spans()
    assert set([span.name for span in spans]) == set(
        [
            "openai.chat",
            "pirate_joke_generator.task",
            "joke_runner.task",
            "joke_manager.workflow",
        ]
    )
    generator_span = next(span for span in spans if span.name == "pirate_joke_generator.task")
    runner_span = next(span for span in spans if span.name == "joke_runner.task")
    manager_span = next(span for span in spans if span.name == "joke_manager.workflow")
    openai_span = next(span for span in spans if span.name == "openai.chat")

    assert openai_span.parent.span_id == generator_span.context.span_id
    assert generator_span.parent.span_id == runner_span.context.span_id
    assert runner_span.parent.span_id == manager_span.context.span_id
    assert openai_span.end_time <= manager_span.end_time


def test_unrelated_entities(exporter):
    @workflow(name="workflow_1")
    def workflow_1():
        return

    @task(name="task_1")
    def task_1():
        return

    workflow_1()
    task_1()

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == ["workflow_1.workflow", "task_1.task"]

    workflow_1_span = spans[0]
    task_1_span = spans[1]

    assert workflow_1_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "workflow_1"
    assert workflow_1_span.attributes[SpanAttributes.TRACELOOP_SPAN_KIND] == "workflow"

    assert task_1_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "task_1"
    assert task_1_span.attributes[SpanAttributes.TRACELOOP_SPAN_KIND] == "task"
    assert task_1_span.parent is None


def test_unserializable_workflow(exporter):
    @task(name="unserializable_task")
    def unserializable_task(obj: object):
        return object()

    @workflow(name="unserializable_workflow")
    def unserializable_workflow(obj: object):
        return unserializable_task(obj)

    unserializable_task(object())

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == ["unserializable_task.task"]


@pytest.mark.asyncio
async def test_unserializable_async_workflow(exporter):
    @task(name="unserializable_task")
    async def unserializable_task(obj: object):
        return object()

    @workflow(name="unserializable_workflow")
    async def unserializable_workflow(obj: object):
        return await unserializable_task(obj)

    await unserializable_task(object())

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == ["unserializable_task.task"]


@pytest.mark.asyncio
async def test_async_generator_workflow(exporter):
    @workflow(name="async generator")
    async def stream_numbers():
        for i in range(3):
            yield i
            await asyncio.sleep(0.1)

    results = []

    async for num in stream_numbers():
        results.append(num)

    spans = exporter.get_finished_spans()

    assert results == [0, 1, 2]
    assert [span.name for span in spans] == ["async generator.workflow"]


def test_sync_workflow_error_handling(exporter):
    """Test error handling in synchronous workflow"""
    error_message = "Intentional workflow test error"

    @workflow(name="failing_workflow")
    def failing_workflow():
        raise ValueError(error_message)

    with pytest.raises(ValueError, match=error_message):
        failing_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 1

    workflow_span = spans[0]
    assert workflow_span.name == "failing_workflow.workflow"
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


@pytest.mark.asyncio
async def test_async_workflow_error_handling(exporter):
    """Test error handling in asynchronous workflow"""
    error_message = "Intentional async workflow test error"

    @workflow(name="failing_async_workflow")
    async def failing_async_workflow():
        await asyncio.sleep(0.1)
        raise ValueError(error_message)

    with pytest.raises(ValueError, match=error_message):
        await failing_async_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 1

    workflow_span = spans[0]
    assert workflow_span.name == "failing_async_workflow.workflow"
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


def test_sync_generator_workflow_error_handling(exporter):
    """Test error handling in synchronous generator workflow"""
    error_message = "Intentional generator workflow test error"

    @workflow(name="failing_generator_workflow")
    def failing_generator_workflow():
        yield 1
        yield 2
        raise ValueError(error_message)

    results = []
    with pytest.raises(ValueError, match=error_message):
        for num in failing_generator_workflow():
            results.append(num)

    spans = exporter.get_finished_spans()
    assert len(spans) == 1

    workflow_span = spans[0]
    assert workflow_span.name == "failing_generator_workflow.workflow"
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description
    assert results == [1, 2]


@pytest.mark.asyncio
async def test_async_generator_workflow_error_handling(exporter):
    """Test error handling in asynchronous generator workflow"""
    error_message = "Intentional async generator workflow test error"

    @workflow(name="failing_async_generator_workflow")
    async def failing_async_generator_workflow():
        yield 1
        await asyncio.sleep(0.1)
        yield 2
        await asyncio.sleep(0.1)
        raise ValueError(error_message)

    results = []
    with pytest.raises(ValueError, match=error_message):
        async for num in failing_async_generator_workflow():
            results.append(num)

    spans = exporter.get_finished_spans()
    assert len(spans) == 1

    workflow_span = spans[0]
    assert workflow_span.name == "failing_async_generator_workflow.workflow"
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description
    assert results == [1, 2]


def test_nested_error_handling(exporter):
    """Test error handling in a workflow that calls a task that raises an error"""
    error_message = "Intentional nested error"

    @task(name="failing_nested_task")
    def failing_nested_task():
        raise ValueError(error_message)

    @workflow(name="parent_workflow")
    def parent_workflow():
        return failing_nested_task()

    with pytest.raises(ValueError, match=error_message):
        parent_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # Check task span
    task_span = next(span for span in spans if span.name == "failing_nested_task.task")
    assert task_span.status.status_code == StatusCode.ERROR
    assert error_message in task_span.status.description

    # Check workflow span
    workflow_span = next(span for span in spans if span.name == "parent_workflow.workflow")
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


@pytest.mark.asyncio
async def test_nested_async_error_handling(exporter):
    """Test error handling in an async workflow that calls an async task that raises an error"""
    error_message = "Intentional nested async error"

    @task(name="failing_nested_async_task")
    async def failing_nested_async_task():
        await asyncio.sleep(0.1)
        raise ValueError(error_message)

    @workflow(name="parent_async_workflow")
    async def parent_async_workflow():
        return await failing_nested_async_task()

    with pytest.raises(ValueError, match=error_message):
        await parent_async_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # Check task span
    task_span = next(span for span in spans if span.name == "failing_nested_async_task.task")
    assert task_span.status.status_code == StatusCode.ERROR
    assert error_message in task_span.status.description

    # Check workflow span
    workflow_span = next(span for span in spans if span.name == "parent_async_workflow.workflow")
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


def test_nested_generator_error_handling(exporter):
    """Test error handling in a workflow that calls a generator task that raises an error"""
    error_message = "Intentional nested generator error"

    @task(name="failing_generator_task")
    def failing_generator_task():
        yield 1
        yield 2
        raise ValueError(error_message)

    @workflow(name="generator_parent_workflow")
    def generator_parent_workflow():
        results = []
        for num in failing_generator_task():
            results.append(num)
        return results

    with pytest.raises(ValueError, match=error_message):
        generator_parent_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # Check task span
    task_span = next(span for span in spans if span.name == "failing_generator_task.task")
    assert task_span.status.status_code == StatusCode.ERROR
    assert error_message in task_span.status.description

    # Check workflow span
    workflow_span = next(span for span in spans if span.name == "generator_parent_workflow.workflow")
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


@pytest.mark.asyncio
async def test_nested_async_generator_error_handling(exporter):
    """Test error handling in an async workflow that calls an async generator task that raises an error"""
    error_message = "Intentional nested async generator error"

    @task(name="failing_async_generator_task")
    async def failing_async_generator_task():
        yield 1
        await asyncio.sleep(0.1)
        yield 2
        await asyncio.sleep(0.1)
        raise ValueError(error_message)

    @workflow(name="async_generator_parent_workflow")
    async def async_generator_parent_workflow():
        results = []
        async for num in failing_async_generator_task():
            results.append(num)
        return results

    with pytest.raises(ValueError, match=error_message):
        await async_generator_parent_workflow()

    spans = exporter.get_finished_spans()
    assert len(spans) == 2

    # Check task span
    task_span = next(span for span in spans if span.name == "failing_async_generator_task.task")
    assert task_span.status.status_code == StatusCode.ERROR
    assert error_message in task_span.status.description

    # Check workflow span
    workflow_span = next(span for span in spans if span.name == "async_generator_parent_workflow.workflow")
    assert workflow_span.status.status_code == StatusCode.ERROR
    assert error_message in workflow_span.status.description


@dataclasses.dataclass
class TestDataClass:
    field1: str
    field2: int


def test_dataclass_serialization_workflow(exporter):
    @task(name="dataclass_task")
    def dataclass_task(data: TestDataClass):
        return data

    @workflow(name="dataclass_workflow")
    def dataclass_workflow(data: TestDataClass):
        return dataclass_task(data)

    data = TestDataClass(field1="value1", field2=123)
    dataclass_workflow(data)

    spans = exporter.get_finished_spans()
    assert [span.name for span in spans] == ["dataclass_task.task", "dataclass_workflow.workflow"]

    task_span = spans[0]
    workflow_span = spans[1]

    assert json.loads(task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_INPUT]) == {
        "args": [{"field1": "value1", "field2": 123}],
        "kwargs": {},
    }
    assert json.loads(task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_OUTPUT]) == {
        "field1": "value1",
        "field2": 123,
    }
    assert task_span.parent.span_id == workflow_span.context.span_id
    assert workflow_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "dataclass_workflow"
    assert task_span.attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] == "dataclass_task"



# generated by datamodel-codegen
# Models for v2/evaluators/execute endpoints from OpenAPI spec
#
# DO NOT EDIT MANUALLY - Regenerate with:
#   ./scripts/generate-models.sh /path/to/swagger.json

from .request import (
    AgentEfficiencyInput,
    AgentEfficiencyRequest,
    AgentFlowQualityConfigRequest,
    AgentFlowQualityInput,
    AgentFlowQualityRequest,
    AgentGoalAccuracyInput,
    AgentGoalAccuracyRequest,
    AgentGoalCompletenessConfigRequest,
    AgentGoalCompletenessInput,
    AgentGoalCompletenessRequest,
    AgentToolErrorDetectorInput,
    AgentToolErrorDetectorRequest,
    AgentToolTrajectoryConfigRequest,
    AgentToolTrajectoryInput,
    AgentToolTrajectoryRequest,
    AnswerCompletenessInput,
    AnswerCompletenessRequest,
    AnswerCorrectnessInput,
    AnswerCorrectnessRequest,
    AnswerRelevancyInput,
    AnswerRelevancyRequest,
    CharCountInput,
    CharCountRatioInput,
    CharCountRatioRequest,
    CharCountRequest,
    ContextRelevanceConfigRequest,
    ContextRelevanceInput,
    ContextRelevanceRequest,
    ConversationQualityInput,
    ConversationQualityRequest,
    FaithfulnessInput,
    FaithfulnessRequest,
    HtmlComparisonInput,
    HtmlComparisonRequest,
    InstructionAdherenceInput,
    InstructionAdherenceRequest,
    IntentChangeInput,
    IntentChangeRequest,
    JSONValidatorConfigRequest,
    JSONValidatorInput,
    JSONValidatorRequest,
    PIIDetectorConfigRequest,
    PIIDetectorInput,
    PIIDetectorRequest,
    PerplexityInput,
    PerplexityRequest,
    PlaceholderRegexConfigRequest,
    PlaceholderRegexInput,
    PlaceholderRegexRequest,
    ProfanityDetectorInput,
    ProfanityDetectorRequest,
    PromptInjectionConfigRequest,
    PromptInjectionInput,
    PromptInjectionRequest,
    PromptPerplexityInput,
    PromptPerplexityRequest,
    RegexValidatorConfigRequest,
    RegexValidatorInput,
    RegexValidatorRequest,
    SQLValidatorInput,
    SQLValidatorRequest,
    SecretsDetectorInput,
    SecretsDetectorRequest,
    SemanticSimilarityInput,
    SemanticSimilarityRequest,
    SexismDetectorConfigRequest,
    SexismDetectorInput,
    SexismDetectorRequest,
    ToneDetectionInput,
    ToneDetectionRequest,
    TopicAdherenceInput,
    TopicAdherenceRequest,
    ToxicityDetectorConfigRequest,
    ToxicityDetectorInput,
    ToxicityDetectorRequest,
    UncertaintyDetectorInput,
    UncertaintyDetectorRequest,
    WordCountInput,
    WordCountRatioInput,
    WordCountRatioRequest,
    WordCountRequest,
)

from .definitions import EvaluatorMadeByTraceloop

from .registry import (
    REQUEST_MODELS,
    RESPONSE_MODELS,
    get_request_model,
    get_response_model,
)

from .response import (
    AgentEfficiencyResponse,
    AgentFlowQualityResponse,
    AgentGoalAccuracyResponse,
    AgentGoalCompletenessResponse,
    AgentToolErrorDetectorResponse,
    AgentToolTrajectoryResponse,
    AnswerCompletenessResponse,
    AnswerCorrectnessResponse,
    AnswerRelevancyResponse,
    CharCountRatioResponse,
    CharCountResponse,
    ContextRelevanceResponse,
    ConversationQualityResponse,
    ErrorResponse,
    FaithfulnessResponse,
    HtmlComparisonResponse,
    InstructionAdherenceResponse,
    IntentChangeResponse,
    JSONValidatorResponse,
    PIIDetectorResponse,
    PerplexityResponse,
    PlaceholderRegexResponse,
    ProfanityDetectorResponse,
    PromptInjectionResponse,
    PromptPerplexityResponse,
    RegexValidatorResponse,
    SQLValidatorResponse,
    SecretsDetectorResponse,
    SemanticSimilarityResponse,
    SexismDetectorResponse,
    ToneDetectionResponse,
    TopicAdherenceResponse,
    ToxicityDetectorResponse,
    UncertaintyDetectorResponse,
    WordCountRatioResponse,
    WordCountResponse,
)

__all__ = [
    # Factory class
    "EvaluatorMadeByTraceloop",
    # Registry functions
    "REQUEST_MODELS",
    "RESPONSE_MODELS",
    "get_request_model",
    "get_response_model",
    # Evaluator request models
    "AgentEfficiencyInput",
    "AgentEfficiencyRequest",
    "AgentFlowQualityConfigRequest",
    "AgentFlowQualityInput",
    "AgentFlowQualityRequest",
    "AgentGoalAccuracyInput",
    "AgentGoalAccuracyRequest",
    "AgentGoalCompletenessConfigRequest",
    "AgentGoalCompletenessInput",
    "AgentGoalCompletenessRequest",
    "AgentToolErrorDetectorInput",
    "AgentToolErrorDetectorRequest",
    "AgentToolTrajectoryConfigRequest",
    "AgentToolTrajectoryInput",
    "AgentToolTrajectoryRequest",
    "AnswerCompletenessInput",
    "AnswerCompletenessRequest",
    "AnswerCorrectnessInput",
    "AnswerCorrectnessRequest",
    "AnswerRelevancyInput",
    "AnswerRelevancyRequest",
    "CharCountInput",
    "CharCountRatioInput",
    "CharCountRatioRequest",
    "CharCountRequest",
    "ContextRelevanceConfigRequest",
    "ContextRelevanceInput",
    "ContextRelevanceRequest",
    "ConversationQualityInput",
    "ConversationQualityRequest",
    "FaithfulnessInput",
    "FaithfulnessRequest",
    "HtmlComparisonInput",
    "HtmlComparisonRequest",
    "InstructionAdherenceInput",
    "InstructionAdherenceRequest",
    "IntentChangeInput",
    "IntentChangeRequest",
    "JSONValidatorConfigRequest",
    "JSONValidatorInput",
    "JSONValidatorRequest",
    "PIIDetectorConfigRequest",
    "PIIDetectorInput",
    "PIIDetectorRequest",
    "PerplexityInput",
    "PerplexityRequest",
    "PlaceholderRegexConfigRequest",
    "PlaceholderRegexInput",
    "PlaceholderRegexRequest",
    "ProfanityDetectorInput",
    "ProfanityDetectorRequest",
    "PromptInjectionConfigRequest",
    "PromptInjectionInput",
    "PromptInjectionRequest",
    "PromptPerplexityInput",
    "PromptPerplexityRequest",
    "RegexValidatorConfigRequest",
    "RegexValidatorInput",
    "RegexValidatorRequest",
    "SQLValidatorInput",
    "SQLValidatorRequest",
    "SecretsDetectorInput",
    "SecretsDetectorRequest",
    "SemanticSimilarityInput",
    "SemanticSimilarityRequest",
    "SexismDetectorConfigRequest",
    "SexismDetectorInput",
    "SexismDetectorRequest",
    "ToneDetectionInput",
    "ToneDetectionRequest",
    "TopicAdherenceInput",
    "TopicAdherenceRequest",
    "ToxicityDetectorConfigRequest",
    "ToxicityDetectorInput",
    "ToxicityDetectorRequest",
    "UncertaintyDetectorInput",
    "UncertaintyDetectorRequest",
    "WordCountInput",
    "WordCountRatioInput",
    "WordCountRatioRequest",
    "WordCountRequest",
    # Evaluator response models
    "AgentEfficiencyResponse",
    "AgentFlowQualityResponse",
    "AgentGoalAccuracyResponse",
    "AgentGoalCompletenessResponse",
    "AgentToolErrorDetectorResponse",
    "AgentToolTrajectoryResponse",
    "AnswerCompletenessResponse",
    "AnswerCorrectnessResponse",
    "AnswerRelevancyResponse",
    "CharCountRatioResponse",
    "CharCountResponse",
    "ContextRelevanceResponse",
    "ConversationQualityResponse",
    "ErrorResponse",
    "FaithfulnessResponse",
    "HtmlComparisonResponse",
    "InstructionAdherenceResponse",
    "IntentChangeResponse",
    "JSONValidatorResponse",
    "PIIDetectorResponse",
    "PerplexityResponse",
    "PlaceholderRegexResponse",
    "ProfanityDetectorResponse",
    "PromptInjectionResponse",
    "PromptPerplexityResponse",
    "RegexValidatorResponse",
    "SQLValidatorResponse",
    "SecretsDetectorResponse",
    "SemanticSimilarityResponse",
    "SexismDetectorResponse",
    "ToneDetectionResponse",
    "TopicAdherenceResponse",
    "ToxicityDetectorResponse",
    "UncertaintyDetectorResponse",
    "WordCountRatioResponse",
    "WordCountResponse",
]
